; WolfARMstein - the GBA raycasting demo
; Mic, 2001/2003
; stabmaster_@hotmail.com | nangijala.no-ip.org
;
; Original Euphoria version be Hollow Horse Software (?)
; Even more original QBasic version by Peter Cooper (?)
;
; I basically translated the code to ARM asm, added some ugly linear shading, etc.
; Note that this program renders the world in a pretty sloppy way, 
; resulting in glitches between walls and stuff like that.
; 

@fsize 32	

@textarea $8000000
@include start.asm					; Boot code



; Convert 8,8,8 colors to 5:5:5 
@macro MakeColor red,green,blue
 	mov r0,red,lsr #3
 	and r1,green,#$F8
 	mov r1,r1,lsl #2
 	orr r0,r0,r1
 	and r1,blue,#$F8 
 	mov r1,r1,lsl #7
 	orr r0,r0,r1
@endm


; Approximate division by 10 as (x*102)/1024
@macro DIV10 dividend,tempreg
	mov	tempreg,#102
	mul	dividend,dividend,tempreg
	mov	dividend,dividend,asr#10
@endm


; Approximate division by 100 as (x*1310)/131072
@macro DIV100 dividend,tempreg
	ldr	tempreg,=1310
	mul	dividend,dividend,tempreg
	mov	dividend,dividend,asr#17
@endm



; ROM code 

start:

	; Copy code from ROM to IWRAM
	mov r0,IWRAM_SECTION
	mov r1,IWRAM_SECTION_ENDS
	sub r2,r1,r0			; # of bytes to move
	mov r1,#$3000000		; Target address
	copy_code
  		ldr r3,[r0],#4
  		str r3,[r1],#4
  		subs r2,r2,#4
	bne copy_code


	mov pc,#$3000000		; Branch to IWRAM


@ltorg					; Dump literal pool


push_msg:
	@incbin push.raw
palette:
	@incbin wstein.pal
main_pal:
	@incbin wstein1.pal
bgimage:
	@incbin wstein.cru
	
@endarea				; End of ROM area



IWRAM_SECTION
@textarea $03000000

b iwram_entry

@include div.asm
@include decrunch.asm


VSync:
  	stmfd sp!,{r0-r1}
  	mov r0,#$4000000
  	add r0,r0,#4			; DISPCNT+4 = DISPSTAT

  	VS_loop1
    		ldrh r1,[r0]
    		ands r1,r1,#1
  	bne VS_loop1

  	VS_loop2
    		ldrh r1,[r0]
    		ands r1,r1,#1
  	beq VS_loop2

  	ldmfd sp!,{r0-r1}
  	mov pc,lr



SetPalette:
; in:
;  r0 = destination
;  r1 = palette
;  r2 = #colors

	stmfd	sp!,{r3}

	SP_loop:
		ldrh	r3,[r1],#2
		strh	r3,[r0],#2
		subs	r2,r2,#1
		bne	SP_loop

	ldmfd	sp!,{r3}
	mov 	pc,lr



SwapBuffers:
	stmfd	sp!,{r0-r3}
 
 	mov     r2,#$04000000
 	ldr	r3,=$FFFFFFEF
       	ldrh    r0,[r2]
       	and 	r1,r0,r3
       	tst 	r0,#$10
       	orreq   r1,r1,#$10
       	strh    r1,[r2]
       	mov 	r0,#$06000000
       	ldrh    r1,[r2]
       	mov     r2,r0
       	tst 	r1,#$10
       	addeq   r0,r0,#$0A000
       	str     r0,[backbuffer]
       
  	ldmfd	sp!,{r0-r3}
  	mov	pc,lr
  	
  	

CheckForWall:
; in:
;  r6 = X (16:16 fp)
;  r7 = Y (16:16 fp)
;
; out:
;  r0 = map color (0-9)
;  Z-flag is set on collision, clear otherwise

	stmfd	sp!,{r1-r5}
	
	mov	r1,r6,asr#16			; X
	mov	r2,r7,asr#16			; Y
	mov	r3,#30
	mov	r4,world
	DIV10	r1,r5
	DIV10	r2,r5
	add	r4,r4,r1
	mul	r2,r2,r3
	add	r4,r4,r2
	ldrb	r0,[r4]
	subs	r0,r0,#'0'
	
	ldmfd	sp!,{r1-r5}
	mov	pc,lr
	

@ltorg


CheckKeys:
	stmfd 	sp!,{r0-r10,lr}

	mov 	r0,#$4000000
	add 	r0,r0,#$130
	
	ldrh	r1,[r0]
	mov	r2,r1
	
	; Turn right
	tst	r1,#$10
	ldreqb	r3,[sinAngle]
	addeq	r3,r3,#4
	streqb	r3,[sinAngle]
	orreq	r2,r2,#$10
	
	; Turn left
	tst	r1,#$20
	ldreqb	r3,[sinAngle]
	addeq	r3,r3,#252			; 256 - 4
	streqb	r3,[sinAngle]
	orreq	r2,r2,#$20
	
	ldrb	r3,[sinAngle]
	ldr	r4,[theta]
	add	r3,r3,r4
	mov	r3,r3,asr#1
	str	r3,[theta]
	and	r3,r3,#$FF
	str	r3,[xTemp]

	; Move forward
	tst	r1,#$40
	bne	no_y_change1
		ldr	r8,[playerX]
		ldr	r9,[playerY]
		mov	r6,r8
		mov 	r7,r9
		mov	r10,sineTable
		ldr	r4,[r10,r3,lsl#2]
		add	r3,r3,#64		; cos(x) = sin(x+PI/2)
		add	r6,r6,r4		
		and	r3,r3,#$FF
		ldr	r4,[r10,r3,lsl#2]
		add	r7,r7,r4
		
		bl CheckForWall
		
		beq	no_wall_1
			mov	r6,r8
			mov	r7,r9
		no_wall_1:
		str	r6,[playerX]
		str	r7,[playerY]
		
		orr	r2,r2,#$40
	no_y_change1:
	
	; Move backward
	tst	r1,#$80
	bne	no_y_change2
		ldr	r8,[playerX]
		ldr	r9,[playerY]
		mov	r6,r8
		mov 	r7,r9
		mov	r10,sineTable
		ldr	r4,[r10,r3,lsl#2]
		add	r3,r3,#64		; cos(x) = sin(x+PI/2)
		sub	r6,r6,r4
		and	r3,r3,#$FF
		ldr	r4,[r10,r3,lsl#2]
		sub	r7,r7,r4
		
		bl CheckForWall
		
		beq	no_wall_2
			mov	r6,r8
			mov	r7,r9
		no_wall_2:
		str	r6,[playerX]
		str	r7,[playerY]
		
		orr	r2,r2,#$80
	no_y_change2:

	mov 	r0,#$4000000
	add 	r0,r0,#$130
	strh	r2,[r0]
	
	ldmfd 	sp!,{r0-r10,lr}
	mov 	pc,lr
	


; A bit over-complicated since all boxes will have a width of 4 and start on
; an even address. But it's generic anyway..
DrawBox:
; in:
;  r0 = x1
;  r1 = y1
;  r2 = x2
;  r3 = y2
;  r4 = color

	stmfd	sp!,{r0-r8}
	
    	mov 	r5,#240
    	mul 	r6,r1,r5
	ldr	r8,[backbuffer]
    	add 	r6,r6,r0
    	add 	r6,r6,r8

    	sub 	r7,r2,r0
    	add 	r7,r7,#1		; Width

    	sub 	r8,r3,r1	
    	add 	r8,r8,#1		; Height

    	rsb 	r5,r7,#240		; 240 - Width

    	mov r3,r4,lsl #8
    	orr r4,r4,r3
    	mov r3,r4,lsl #16
    	orr r4,r4,r3			; Color

    	box_y_loop:
      		mov r3,r7
		
		; Handle the case where the address is odd
		tst	r7,#1
		ldrneh	r0,[r6,-1]
		andne	r2,r4,#$FF00
		andne	r0,r0,#$00FF
		addne	r6,r6,#1
		orrne	r0,r0,r2
		subne	r3,r3,#1
		strneh	r0,[r6,-2]
		
		; Draw groups of 4 pixels
      		mov r1,r3,lsr #2
      		box_x_loop:
        		str r4,[r6],#4
        		subs r1,r1,#1
      		bne box_x_loop

		; Draw 2 pixels (if any remaining)      		
      		tst	r3,#2
      		strneh	r4,[r6],#2

		; Draw the last pixel (if any remaining)      		
      		tst	r3,#1
      		ldrneh	r2,[r6]
      		andne	r0,r4,#$FF
      		andne	r2,r2,#$FF00
      		orrne	r2,r2,r0
		strneh	r2,[r6],#1
		
      		add 	r6,r6,r5
      		subs 	r8,r8,#1
    	bne box_y_loop

	ldmfd	sp!,{r0-r8}
	mov	pc,lr
	

@ltorg


FadeIn:
; in:
;  r0 = source palette
;  r1 = target palette

	stmfd	sp!,{r2-r7}
	
	mov	r2,#256
	FI_loop:
		ldrh	r3,[r0]
		ldrh	r4,[r1],#2
		
		mov	r7,#0
		
		and	r5,r3,#$1F
		and	r6,r4,#$1F
		cmp	r5,r6
		addlt	r5,r5,#1
		subgt	r5,r5,#1
		orr	r7,r7,r5
	
		mov	r3,r3,lsr#5
		mov	r4,r4,lsr#5
		and	r5,r3,#$1F
		and	r6,r4,#$1F
		cmp	r5,r6
		addlt	r5,r5,#1
		subgt	r5,r5,#1
		mov	r5,r5,lsl#5
		orr	r7,r7,r5

		mov	r3,r3,lsr#5
		mov	r4,r4,lsr#5
		and	r5,r3,#$1F
		and	r6,r4,#$1F
		cmp	r5,r6
		addlt	r5,r5,#1
		subgt	r5,r5,#1
		mov	r5,r5,lsl#10
		orr	r7,r7,r5
		
		strh	r7,[r0],#2
		
		subs	r2,r2,#1
	bne	FI_loop	
	
	ldmfd	sp!,{r2-r7}
	mov pc,lr
	



Delay:
; in:
;  r0 = number of delays

	stmfd	sp!,{r1-r2,lr}
    	
    	delay_loop:
		bl	VSync
      		subs 	r0,r0,#1
    	bne delay_loop
	
	ldmfd 	sp!,{r1-r2,lr}
	mov	pc,lr





; Main 
iwram_entry

bl	VSync

; Set mode 4 and enable BG2
mov 	r0,#$0400
add 	r0,r0,#4
mov 	r1,#$4000000
str 	r0,[r1]			

; Intro

mov	r0,#4
bl	Delay

bl	VSync

mov	r0,#$05000000
mov	r1,work_pal
mov	r2,#256
bl	SetPalette

; Decompress the background image to VRAM
mov	r0,bgimage
ldr	r1,=$0600A000
mov	r2,#2
bl	DeCrunch

bl	VSync
bl	SwapBuffers

mov	r0,#$06000000
ldr	r1,=$0600A000
ldr	r2,=19200
copy_buffer:
	ldrh	r3,[r1],#2
	strh	r3,[r0],#2
	subs	r2,r2,#1
bne	copy_buffer


; Overlay the "Push Start button" message
ldr	r0,=$060070C8
mov	r1,push_msg
mov	r2,#12
overlay_msg_y:
	mov	r3,#48
	overlay_msg_x:
		ldrh	r4,[r0]
		ldrh	r5,[r1],#2

		ands	r6,r5,#$00FF
		andne	r4,r4,#$FF00
		orrne	r4,r4,#$60

		ands	r5,r5,#$FF00
		andne	r4,r4,#$00FF
		orrne	r4,r4,#$6000

		strh	r4,[r0],#2
		
		subs	r3,r3,#1
	bne overlay_msg_x
	add	r0,r0,#144
	subs	r2,r2,#1
bne	overlay_msg_y


mov	r5,#32
fade:
	mov	r0,work_pal
	mov	r1,palette
	bl	FadeIn
	bl	VSync
	bl	VSync
	bl	VSync
	bl	VSync
	mov	r0,#$05000000
	mov	r1,work_pal
	mov	r2,#256
	bl	SetPalette
	subs	r5,r5,#1
bne 	fade



mov 	r0,#$04000000
orr	r0,r0,#$130
mov	r4,#1
mov	r5,#0
wait_for_keypress:
	ldrh 	r1,[r0]
	tst	r1,#1
	beq	end_wait
	ldrh 	r1,[r0]
	tst	r1,#2
	beq	end_wait
	ldrh 	r1,[r0]
	tst	r1,#8
	beq	end_wait
	
	bl	VSync
	
	subs	r4,r4,#1
	bne	wait_for_keypress
	bl	SwapBuffers
	eors	r5,r5,#1
	moveq	r4,#18
	movne	r4,#32
b wait_for_keypress

end_wait:
orr	r1,r1,#$FF
strh	r1,[r0]


bl	SwapBuffers

mov	r0,#$05000000
mov	r1,main_pal
mov	r2,#256
bl	SetPalette
  

; Main loop
forever:

	; Draw sky and ground
	ldr	r0,[backbuffer]
	ldr	r1,=2526451350				; Color 150
	ldr	r2,=3452816845				; Color 205
	ldr	r3,=$4B00
	mov	r4,r3,lsr#2
	set_bg_color:
		str	r1,[r0]
		str	r2,[r0,r3]
		subs	r4,r4,#1
		add	r0,r0,#4
	bne set_bg_color
	
	
	ldrb	r0,[sinAngle]
	mov	r10,r0					; Loop counter (i)
	add	r0,r0,#60				; 60 degrees FOV
	str	r0,[theta]

	FOV_loop:
		mov	r0,sineTable
		and	r1,r10,#$FF
		ldr	r2,[r0,r1,lsl#2]		; r2 = sin(i) (deltaX)
		add	r1,r1,#64
		mov	r2,r2,asr#1
		and	r1,r1,#$FF
		ldr	r6,[playerX]
		ldr	r3,[r0,r1,lsl#2]		; r3 = cos(i) (deltaY)
		ldr	r7,[playerY]
		mov	r8,#0				; Distance to intersection
		mov	r3,r3,asr#1

		orrs	r8,r8,r8			; Just to make the first iteration work
		
		find_intersection:
			bne	found_intersection
			add	r6,r6,r2		; x += deltaX		
			add	r7,r7,r3		; y += deltaY
			add	r8,r8,#1		; distance++
			
			bl	CheckForWall
		b find_intersection
		
		found_intersection:
		mov	r7,r0				; Color of the cell (1-9)

		
		; Calculate the height of the cell
		ldr	r0,=700
		mov	r1,r8
		bl	FDIV
		mov	r9,r0				; height = 700/distance
		cmp	r9,#79
		movgt	r9,#79				; if (height>79) height = 79 

		; Calculate the shade based on the distance
		mov	r4,#15
		mul	r4,r4,r8
		DIV100	r4,r2				; shade = (distance*15)/100
		cmp	r4,#15
		movgt	r4,#15				; if (shade>15) shade = 15
		
		ldrb	r2,[sinAngle]
		rsb	r1,r9,#79			; y1 = 79-height
		sub	r0,r10,r2			; x1 = i-sinAngle
		add	r3,r9,#80			; y2 = height+80
		mov	r7,r7,lsl#4	
		mov	r0,r0,lsl#2			; x1 <<= 2
		sub	r4,r7,r4
		add	r2,r0,#3			; x2 = x0 + 3
		bl	DrawBox
		
		ldr	r0,[theta]
		add	r10,r10,#1			; i++
		cmp	r10,r0
	blt	FOV_loop

  	bl	VSync
  	bl	SwapBuffers	
	bl	CheckKeys
	
b forever


@ltorg			; Dump literal pool


; RW DATA

sinAngle	@dcb 0
xTemp		@dcb 0
theta		@dcd 0
playerX		@dcd $1D0000
playerY		@dcd $130000
backbuffer	@dcd $06000000


; Sine table, 256 elements in 16:16 fixed point format	
sineTable:
	@incbin sin1616.bin

; The world, 30x20	
world:
	@dcb   "191919191991991199119919199199"
	@dcb   "900000000070007000011050000001"
	@dcb   "100000000070007104110500005009"
	@dcb   "900100500057077000110500003001"
	@dcb   "100004000070507011050000030009"
	@dcb   "903000000040707011050000002041"
	@dcb   "100007800040002001105000000009"
	@dcb   "905008700070007001105000066001"
	@dcb   "100000000000000011050000040009"
	@dcb   "108888800001105000000000070009"
	@dcb   "108800080005500110500000070009"
	@dcb   "108000080005500011050000000009"
	@dcb   "108000080005500110500000433009"
	@dcb   "108070000000001105000000000509"
	@dcb   "108000000000440001105000000509"
	@dcb   "108000560011050000000000230509"
	@dcb   "100000200000003200110500000609"
	@dcb   "100033300000000441105000886009"
	@dcb   "100000000001105000000008800009"
	@dcb   "919191991191191991919191919111"

work_pal:	

@endarea
IWRAM_SECTION_ENDS

